@@ -9,28 +9,27 @@
 
 
 #include "stdafx.h"
-#include "Loaders.h"
 #include "S3MTools.h"
-#include "../common/StringFixer.h"
+#include "Loaders.h"
+#include "../common/mptStringBuffer.h"
 
 
 OPENMPT_NAMESPACE_BEGIN
 
-
 // Convert an S3M sample header to OpenMPT's internal sample header.
 void S3MSampleHeader::ConvertToMPT(ModSample &mptSmp, bool isST3) const
 {
 	mptSmp.Initialize(MOD_TYPE_S3M);
-	mpt::String::Read<mpt::String::maybeNullTerminated>(mptSmp.filename, filename);
+	mptSmp.filename = mpt::String::ReadBuf(mpt::String::maybeNullTerminated, filename);
 
-	if((sampleType == typePCM || sampleType == typeNone) && !memcmp(magic, "SCRS", 4))
+	if(sampleType == typePCM || sampleType == typeNone)
 	{
 		// Sample Length and Loops
 		if(sampleType == typePCM)
 		{
 			mptSmp.nLength = length;
-			mptSmp.nLoopStart = MIN(loopStart, mptSmp.nLength - 1);
-			mptSmp.nLoopEnd = MIN(loopEnd, mptSmp.nLength);
+			mptSmp.nLoopStart = std::min(static_cast<SmpLength>(loopStart), mptSmp.nLength - 1);
+			mptSmp.nLoopEnd = std::min(static_cast<SmpLength>(loopEnd), mptSmp.nLength);
 			mptSmp.uFlags.set(CHN_LOOP, (flags & smpLoop) != 0);
 		}
 
@@ -39,21 +38,34 @@ void S3MSampleHeader::ConvertToMPT(ModSample &mptSmp, bool isST3) const
 			mptSmp.nLoopStart = mptSmp.nLoopEnd = 0;
 			mptSmp.uFlags.reset();
 		}
-
-		// Volume / Panning
-		mptSmp.nVolume = MIN(defaultVolume, 64) * 4;
-
-		// C-5 frequency
-		mptSmp.nC5Speed = c5speed;
-		// ST3 clamps samples to 65536 Hz
-		if(isST3)
-			LimitMax(mptSmp.nC5Speed, uint16_max);
-
-		if(mptSmp.nC5Speed == 0)
-			mptSmp.nC5Speed = 8363;
-		else if(mptSmp.nC5Speed < 1024)
-			mptSmp.nC5Speed = 1024;
+	} else if(sampleType == typeAdMel)
+	{
+		OPLPatch patch;
+		std::memcpy(patch.data() + 0, mpt::as_raw_memory(length).data(), 4);
+		std::memcpy(patch.data() + 4, mpt::as_raw_memory(loopStart).data(), 4);
+		std::memcpy(patch.data() + 8, mpt::as_raw_memory(loopEnd).data(), 4);
+		mptSmp.SetAdlib(true, patch);
 	}
+
+	// Volume / Panning
+	mptSmp.nVolume = std::min(defaultVolume.get(), uint8(64)) * 4;
+
+	// C-5 frequency
+	mptSmp.nC5Speed = c5speed;
+	if(isST3)
+	{
+		// ST3 ignores or clamps the high 16 bits depending on the instrument type
+		if(sampleType == typeAdMel)
+			mptSmp.nC5Speed &= 0xFFFF;
+		else
+			LimitMax(mptSmp.nC5Speed, uint16_max);
+	}
+
+	if(mptSmp.nC5Speed == 0)
+		mptSmp.nC5Speed = 8363;
+	else if(mptSmp.nC5Speed < 1024)
+		mptSmp.nC5Speed = 1024;
+
 }
 
 
@@ -61,14 +73,22 @@ void S3MSampleHeader::ConvertToMPT(ModSample &mptSmp, bool isST3) const
 SmpLength S3MSampleHeader::ConvertToS3M(const ModSample &mptSmp)
 {
 	SmpLength smpLength = 0;
-	mpt::String::Write<mpt::String::maybeNullTerminated>(filename, mptSmp.filename);
+	mpt::String::WriteBuf(mpt::String::maybeNullTerminated, filename) = mptSmp.filename;
+	memcpy(magic, "SCRS", 4);
 
-	if(mptSmp.pSample != nullptr)
+	if(mptSmp.uFlags[CHN_ADLIB])
+	{
+		memcpy(magic, "SCRI", 4);
+		sampleType = typeAdMel;
+		std::memcpy(mpt::as_raw_memory(length   ).data(), mptSmp.adlib.data() + 0, 4);
+		std::memcpy(mpt::as_raw_memory(loopStart).data(), mptSmp.adlib.data() + 4, 4);
+		std::memcpy(mpt::as_raw_memory(loopEnd  ).data(), mptSmp.adlib.data() + 8, 4);
+	} else if(mptSmp.HasSampleData())
 	{
 		sampleType = typePCM;
-		length = static_cast<uint32>(MIN(mptSmp.nLength, uint32_max));
-		loopStart = static_cast<uint32>(MIN(mptSmp.nLoopStart, uint32_max));
-		loopEnd = static_cast<uint32>(MIN(mptSmp.nLoopEnd, uint32_max));
+		length = mpt::saturate_cast<uint32>(mptSmp.nLength);
+		loopStart = mpt::saturate_cast<uint32>(mptSmp.nLoopStart);
+		loopEnd = mpt::saturate_cast<uint32>(mptSmp.nLoopEnd);
 
 		smpLength = length;
 
@@ -86,7 +106,7 @@ SmpLength S3MSampleHeader::ConvertToS3M(const ModSample &mptSmp)
 		sampleType = typeNone;
 	}
 
-	defaultVolume = static_cast<uint8>(MIN(mptSmp.nVolume / 4, 64));
+	defaultVolume = static_cast<uint8>(std::min(static_cast<uint16>(mptSmp.nVolume / 4), uint16(64)));
 	if(mptSmp.nC5Speed != 0)
 	{
 		c5speed = mptSmp.nC5Speed;
@@ -94,7 +114,6 @@ SmpLength S3MSampleHeader::ConvertToS3M(const ModSample &mptSmp)
 	{
 		c5speed = ModSample::TransposeToFrequency(mptSmp.RelativeTone, mptSmp.nFineTune);
 	}
-	memcpy(magic, "SCRS", 4);
 
 	return smpLength;
 }
@@ -118,4 +137,11 @@ SampleIO S3MSampleHeader::GetSampleFormat(bool signedSamples) const
 }
 
 
+// Calculate the sample position in file
+uint32 S3MSampleHeader::GetSampleOffset() const
+{
+	return (dataPointer[1] << 4) | (dataPointer[2] << 12) | (dataPointer[0] << 20);
+}
+
+
 OPENMPT_NAMESPACE_END
