@@ -19,9 +19,6 @@
 #include "../common/mptFileType.h"
 #include "../common/mptRandom.h"
 #include "../common/version.h"
-#include <vector>
-#include <bitset>
-#include <set>
 #include "Snd_defs.h"
 #include "tuningbase.h"
 #include "MIDIMacros.h"
@@ -44,21 +41,26 @@
 #include "../sounddsp/EQ.h"
 #endif
 
-#include "modcommand.h"
-#include "ModSample.h"
-#include "ModInstrument.h"
-#include "ModChannel.h"
-#include "plugins/PluginStructs.h"
-#include "RowVisitor.h"
 #include "Message.h"
+#include "ModChannel.h"
+#include "modcommand.h"
+#include "ModInstrument.h"
+#include "ModSample.h"
+#include "ModSequence.h"
 #include "pattern.h"
 #include "patternContainer.h"
-#include "ModSequence.h"
+#include "PlayState.h"
+#include "plugins/PluginStructs.h"
+#include "RowVisitor.h"
 
 #include "mpt/audio/span.hpp"
 
 #include "../common/FileReaderFwd.h"
 
+#include <vector>
+#include <bitset>
+#include <set>
+
 
 OPENMPT_NAMESPACE_BEGIN
 
@@ -66,17 +68,6 @@ OPENMPT_NAMESPACE_BEGIN
 bool SettingCacheCompleteFileBeforeLoading();
 
 
-// -----------------------------------------------------------------------------
-// MODULAR ModInstrument FIELD ACCESS : body content in InstrumentExtensions.cpp
-// -----------------------------------------------------------------------------
-#ifndef MODPLUG_NO_FILESAVE
-void WriteInstrumentHeaderStructOrField(ModInstrument * input, std::ostream &file, uint32 only_this_code = -1 /* -1 for all */, uint16 fixedsize = 0);
-#endif // !MODPLUG_NO_FILESAVE
-bool ReadInstrumentHeaderField(ModInstrument * input, uint32 fcode, uint16 fsize, FileReader &file);
-// --------------------------------------------------------------------------------------------
-// --------------------------------------------------------------------------------------------
-
-
 // Sample decompression routines in format-specific source files
 void AMSUnpack(mpt::const_byte_span source, mpt::byte_span dest, int8 packCharacter);
 uintptr_t DMFUnpack(FileReader &file, uint8 *psample, uint32 maxlen);
@@ -107,14 +98,14 @@ struct PatternCuePoint
 // Return values for GetLength()
 struct GetLengthType
 {
-	double     duration = 0.0;                 // Total time in seconds
-	ROWINDEX   lastRow = ROWINDEX_INVALID;     // Last parsed row (if no target is specified, this is the first row that is parsed twice, i.e. not the *last* played order)
-	ROWINDEX   endRow = ROWINDEX_INVALID;      // Last row before module loops (UNDEFINED if a target is specified)
-	ROWINDEX   startRow = 0;                   // First row of parsed subsong
-	ORDERINDEX lastOrder = ORDERINDEX_INVALID; // Last parsed order (see lastRow remark)
-	ORDERINDEX endOrder = ORDERINDEX_INVALID;  // Last order before module loops (UNDEFINED if a target is specified)
-	ORDERINDEX startOrder = 0;                 // First order of parsed subsong
-	bool       targetReached = false;          // True if the specified order/row combination or duration has been reached while going through the module
+	double     duration = 0.0;                     // Total time in seconds
+	ROWINDEX   restartRow = ROWINDEX_INVALID;      // First row to play after module loops (or last parsed row if target is specified; equal to target if it was found)
+	ROWINDEX   endRow = ROWINDEX_INVALID;          // Last row played before module loops (UNDEFINED if a target is specified)
+	ROWINDEX   startRow = 0;                       // First row of parsed subsong
+	ORDERINDEX restartOrder = ORDERINDEX_INVALID;  // First row to play after module loops (see restartRow remark)
+	ORDERINDEX endOrder = ORDERINDEX_INVALID;      // Last order played before module loops (UNDEFINED if a target is specified)
+	ORDERINDEX startOrder = 0;                     // First order of parsed subsong
+	bool       targetReached = false;              // True if the specified order/row combination or duration has been reached while going through the module
 };
 
 
@@ -233,6 +224,7 @@ class CTuningCollection;
 using CTuningCollection = Tuning::CTuningCollection;
 struct CModSpecifications;
 class OPL;
+class PlaybackTest;
 class CModDoc;
 
 
@@ -254,6 +246,19 @@ struct FileHistory
 	{
 		return loadDate != mpt::Date::AnyGregorian{};
 	}
+
+	bool operator==(const FileHistory &other) const noexcept
+	{
+		return std::tie(loadDate, openTime) == std::tie(other.loadDate, other.openTime);
+	}
+	bool operator!=(const FileHistory &other) const noexcept
+	{
+		return std::tie(loadDate, openTime) != std::tie(other.loadDate, other.openTime);
+	}
+	bool operator<(const FileHistory &other) const noexcept
+	{
+		return std::tie(loadDate, openTime) < std::tie(other.loadDate, other.openTime);
+	}
 };
 
 
@@ -360,9 +365,34 @@ public:
 using NoteName = mpt::uchar[4];
 
 
+struct PlaybackTestSettings
+{
+	uint32 mixingFreq = 48000;
+	uint32 outputChannels = 2;
+	uint32 mixerChannels = MAX_CHANNELS;
+	ResamplingMode srcMode = SRCMODE_CUBIC;
+	void Sanitize()
+	{
+		if(mixingFreq < 1000)
+		{
+			mixingFreq = 48000;
+		}
+		if(outputChannels != 1 && outputChannels != 2 && outputChannels != 4)
+		{
+			outputChannels = 2;
+		}
+		if(mixerChannels < 1)
+		{
+			mixerChannels = MAX_CHANNELS;
+		}
+	}
+};
+
+
 class CSoundFile
 {
 	friend class GetLengthMemory;
+	friend class MIDIMacroParser;
 
 public:
 #ifdef MODPLUG_TRACKER
@@ -379,7 +409,6 @@ public:
 	//Tuning-->
 public:
 	static std::unique_ptr<CTuning> CreateTuning12TET(const mpt::ustring &name);
-	static CTuning *GetDefaultTuning() {return nullptr;}
 	CTuningCollection& GetTuneSpecificTunings() {return *m_pTuningsTuneSpecific;}
 
 	mpt::ustring GetNoteName(const ModCommand::NOTE note, const INSTRUMENTINDEX inst, const NoteName *noteNames = nullptr) const;
@@ -403,9 +432,6 @@ private:
 private:
 	CTuningCollection* m_pTuningsTuneSpecific = nullptr;
 
-private: //Misc private methods.
-	static void SetModSpecsPointer(const CModSpecifications* &pModSpecs, const MODTYPE type);
-
 private: //Misc data
 	const CModSpecifications *m_pModSpecs;
 
@@ -443,29 +469,24 @@ public:
 	BitCrush m_BitCrush;
 #endif
 
-	using samplecount_t = uint32; // Number of rendered samples
-
 	static constexpr uint32 TICKS_ROW_FINISHED = uint32_max - 1u;
 
-public:	// for Editing
+private:
 #ifdef MODPLUG_TRACKER
-	CModDoc *m_pModDoc = nullptr; // Can be a null pointer for example when previewing samples from the treeview.
+	CModDoc *m_pModDoc = nullptr;  // Can be a null pointer for example when previewing samples from the treeview.
 #endif // MODPLUG_TRACKER
 	Enum<MODTYPE> m_nType;
-private:
 	ModContainerType m_ContainerType = ModContainerType::None;
 public:
-	CHANNELINDEX m_nChannels = 0;
 	SAMPLEINDEX m_nSamples = 0;
 	INSTRUMENTINDEX m_nInstruments = 0;
-	uint32 m_nDefaultSpeed, m_nDefaultGlobalVolume;
-	TEMPO m_nDefaultTempo;
+	uint32 m_nDefaultGlobalVolume;
 	FlagSet<SongFlags> m_SongFlags;
 	CHANNELINDEX m_nMixChannels = 0;
 private:
 	CHANNELINDEX m_nMixStat;
 public:
-	ROWINDEX m_nDefaultRowsPerBeat, m_nDefaultRowsPerMeasure;	// default rows per beat and measure for this module
+	ROWINDEX m_nDefaultRowsPerBeat, m_nDefaultRowsPerMeasure;  // default rows per beat and measure for this module
 	TempoMode m_nTempoMode = TempoMode::Classic;
 
 #ifdef MODPLUG_TRACKER
@@ -495,14 +516,15 @@ public:
 
 	ResamplingMode m_nResampling; // Resampling mode (if overriding the globally set resampling)
 	int32 m_nRepeatCount = 0;     // -1 means repeat infinitely.
-	ORDERINDEX m_nMaxOrderPosition;
-	ModChannelSettings ChnSettings[MAX_BASECHANNELS];  // Initial channels settings
+	ORDERINDEX m_restartOverridePos = 0, m_maxOrderPosition = 0;
+	std::vector<ModChannelSettings> ChnSettings;  // Initial channels settings
 	CPatternContainer Patterns;
 	ModSequenceSet Order;  // Pattern sequences (order lists)
 protected:
 	ModSample Samples[MAX_SAMPLES];
 public:
 	ModInstrument *Instruments[MAX_INSTRUMENTS];  // Instrument Headers
+	InstrumentSynth::Events m_globalScript;
 	MIDIMacroConfig m_MidiCfg;                    // MIDI Macro config table
 #ifndef NO_PLUGINS
 	std::array<SNDMIXPLUGIN, MAX_MIXPLUGINS> m_MixPlugins;  // Mix plugins
@@ -527,84 +549,6 @@ protected:
 	MixLevels m_nMixLevels;
 
 public:
-	struct PlayState
-	{
-		friend class CSoundFile;
-
-	public:
-		samplecount_t m_lTotalSampleCount = 0;  // Total number of rendered samples
-	protected:
-		samplecount_t m_nBufferCount = 0;  // Remaining number samples to render for this tick
-		double m_dBufferDiff = 0.0;        // Modern tempo rounding error compensation
-
-	public:
-		uint32 m_nTickCount = 0;  // Current tick being processed
-	protected:
-		uint32 m_nPatternDelay = 0;  // Pattern delay (rows)
-		uint32 m_nFrameDelay = 0;    // Fine pattern delay (ticks)
-	public:
-		uint32 m_nSamplesPerTick = 0;
-		ROWINDEX m_nCurrentRowsPerBeat = 0;     // Current time signature
-		ROWINDEX m_nCurrentRowsPerMeasure = 0;  // Current time signature
-		uint32 m_nMusicSpeed = 0;               // Current speed
-		TEMPO m_nMusicTempo;                    // Current tempo
-
-		// Playback position
-		ROWINDEX m_nRow = 0;      // Current row being processed
-		ROWINDEX m_nNextRow = 0;  // Next row to process
-	protected:
-		ROWINDEX m_nextPatStartRow = 0;  // For FT2's E60 bug
-		ROWINDEX m_breakRow = 0;          // Candidate target row for pattern break
-		ROWINDEX m_patLoopRow = 0;        // Candidate target row for pattern loop
-		ORDERINDEX m_posJump = 0;         // Candidate target order for position jump
-
-	public:
-		PATTERNINDEX m_nPattern = 0;                     // Current pattern being processed
-		ORDERINDEX m_nCurrentOrder = 0;                  // Current order being processed
-		ORDERINDEX m_nNextOrder = 0;                     // Next order to process
-		ORDERINDEX m_nSeqOverride = ORDERINDEX_INVALID;  // Queued order to be processed next, regardless of what order would normally follow
-
-		// Global volume
-	public:
-		int32 m_nGlobalVolume = MAX_GLOBAL_VOLUME;  // Current global volume (0...MAX_GLOBAL_VOLUME)
-	protected:
-		int32 m_nSamplesToGlobalVolRampDest = 0, m_nGlobalVolumeRampAmount = 0,
-		      m_nGlobalVolumeDestination = 0;     // Global volume ramping
-		int32 m_lHighResRampingGlobalVolume = 0;  // Global volume ramping
-
-	public:
-		bool m_bPositionChanged = true; // Report to plugins that we jumped around in the module
-
-	public:
-		CHANNELINDEX ChnMix[MAX_CHANNELS]; // Index of channels in Chn to be actually mixed
-		ModChannel Chn[MAX_CHANNELS];      // Mixing channels... First m_nChannels channels are master channels (i.e. they are never NNA channels)!
-
-		struct MIDIMacroEvaluationResults
-		{
-			std::map<PLUGINDEX, float> pluginDryWetRatio;
-			std::map<std::pair<PLUGINDEX, PlugParamIndex>, PlugParamValue> pluginParameter;
-		};
-
-		std::vector<uint8> m_midiMacroScratchSpace;
-		std::optional<MIDIMacroEvaluationResults> m_midiMacroEvaluationResults;
-
-	public:
-		PlayState();
-
-		void ResetGlobalVolumeRamping()
-		{
-			m_lHighResRampingGlobalVolume = m_nGlobalVolume << VOLUMERAMPPRECISION;
-			m_nGlobalVolumeDestination = m_nGlobalVolume;
-			m_nSamplesToGlobalVolRampDest = 0;
-			m_nGlobalVolumeRampAmount = 0;
-		}
-
-		constexpr uint32 TicksOnRow() const noexcept
-		{
-			return (m_nMusicSpeed + m_nFrameDelay) * std::max(m_nPatternDelay, uint32(1));
-		}
-	};
-
 	PlayState m_PlayState;
 
 protected:
@@ -631,11 +575,9 @@ private:
 #endif // MODPLUG_TRACKER
 
 public:
-#ifdef LIBOPENMPT_BUILD
-#ifndef NO_PLUGINS
+#if defined(LIBOPENMPT_BUILD) && !defined(NO_PLUGINS)
 	std::unique_ptr<CVstPluginManager> m_PluginManager;
 #endif
-#endif
 
 public:
 	std::string m_songName;
@@ -686,13 +628,13 @@ public:
 
 	enum ModLoadingFlags
 	{
-		onlyVerifyHeader   = 0x00,
 		loadPatternData    = 0x01, // If unset, advise loaders to not process any pattern data (if possible)
 		loadSampleData     = 0x02, // If unset, advise loaders to not process any sample data (if possible)
-		loadPluginData     = 0x04, // If unset, plugin data is not loaded (and as a consequence, plugins are not instanciated).
-		loadPluginInstance = 0x08, // If unset, plugins are not instanciated.
+		loadPluginData     = 0x04, // If unset, plugin data is not loaded (and as a consequence, plugins are not instantiated).
+		loadPluginInstance = 0x08, // If unset, plugins are not instantiated.
 		skipContainer      = 0x10,
 		skipModules        = 0x20,
+		onlyVerifyHeader   = 0x40, // Do not combine with other flags!
 
 		// Shortcuts
 		loadCompleteModule = loadSampleData | loadPatternData | loadPluginData | loadPluginInstance,
@@ -731,6 +673,7 @@ public:
 	CModDoc *GetpModDoc() const noexcept { return m_pModDoc; }
 #endif  // MODPLUG_TRACKER
 
+	void Create(MODTYPE type, CHANNELINDEX numChannels, CModDoc *pModDoc = nullptr);
 	bool Create(FileReader file, ModLoadingFlags loadFlags = loadCompleteModule, CModDoc *pModDoc = nullptr);
 private:
 	bool CreateInternal(FileReader file, ModLoadingFlags loadFlags);
@@ -772,7 +715,7 @@ public:
 	constexpr SAMPLEINDEX GetNumSamples() const noexcept { return m_nSamples; }
 	constexpr PATTERNINDEX GetCurrentPattern() const noexcept { return m_PlayState.m_nPattern; }
 	constexpr ORDERINDEX GetCurrentOrder() const noexcept { return m_PlayState.m_nCurrentOrder; }
-	constexpr CHANNELINDEX GetNumChannels() const noexcept { return m_nChannels; }
+	MPT_FORCEINLINE CHANNELINDEX GetNumChannels() const noexcept { return static_cast<CHANNELINDEX>(ChnSettings.size()); }
 
 	constexpr bool CanAddMoreSamples(SAMPLEINDEX amount = 1) const noexcept { return (amount < MAX_SAMPLES) && m_nSamples < (MAX_SAMPLES - amount); }
 	constexpr bool CanAddMoreInstruments(INSTRUMENTINDEX amount = 1) const noexcept { return (amount < MAX_INSTRUMENTS) && m_nInstruments < (MAX_INSTRUMENTS - amount); }
@@ -786,11 +729,25 @@ public:
 	static ChannelFlags GetChannelMuteFlag();
 
 #ifdef MODPLUG_TRACKER
-	void PatternTranstionChnSolo(const CHANNELINDEX chnIndex);
+	void PatternTranstionChnSolo(const CHANNELINDEX first, const CHANNELINDEX last);
 	void PatternTransitionChnUnmuteAll();
 
 protected:
-	void HandlePatternTransitionEvents();
+	void HandleRowTransitionEvents(bool nextPattern);
+
+	const ModSample *m_metronomeMeasure = nullptr;
+	const ModSample *m_metronomeBeat = nullptr;
+	ModChannel m_metronomeChn{};
+
+public:
+	void SetMetronomeSamples(const ModSample *measure, const ModSample *beat)
+	{
+		m_metronomeMeasure = measure;
+		m_metronomeBeat = beat;
+		m_metronomeChn.pModSample = nullptr;
+		m_metronomeChn.pCurrentSample = nullptr;
+	}
+	constexpr bool IsMetronomeEnabled() const noexcept { return m_metronomeMeasure || m_metronomeBeat; }
 #endif  // MODPLUG_TRACKER
 
 public:
@@ -819,10 +776,9 @@ public:
 	// A repeat count value of -1 means infinite loop
 	void SetRepeatCount(int n) { m_nRepeatCount = n; }
 	int GetRepeatCount() const { return m_nRepeatCount; }
-	bool IsPaused() const { return m_SongFlags[SONG_PAUSED | SONG_STEP]; }	// Added SONG_STEP as it seems to be desirable in most cases to check for this as well.
+	bool IsPaused() const { return m_PlayState.m_flags[SONG_PAUSED | SONG_STEP]; }	// Added SONG_STEP as it seems to be desirable in most cases to check for this as well.
 	void LoopPattern(PATTERNINDEX nPat, ROWINDEX nRow = 0);
 
-	bool InitChannel(CHANNELINDEX nChn);
 	void InitAmigaResampler();
 
 	void InitOPL();
@@ -844,6 +800,7 @@ public:
 	static ProbeResult ProbeFileHeaderAMS(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderAMS2(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderC67(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderCBA(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderDBM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderDTM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderDIGI(MemoryFileReader file, const uint64 *pfilesize);
@@ -851,18 +808,24 @@ public:
 	static ProbeResult ProbeFileHeaderDSM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderDSm(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderDSym(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderETX(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderFAR(MemoryFileReader file, const uint64 *pfilesize);
-	static ProbeResult ProbeFileHeaderFMT(MemoryFileReader file, const uint64* pfilesize);
+	static ProbeResult ProbeFileHeaderFC(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderFMT(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderFTM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderGDM(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderGMC(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderGT2(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderGTK(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderICE(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderIMF(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderIMS(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderIT(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderITP(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderJ2B(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderKRIS(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderMUS_KM(MemoryFileReader file, const uint64 *pfilesize);
-	static ProbeResult ProbeFileHeaderM15(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderSTK(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderMDL(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderMED(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderMO3(MemoryFileReader file, const uint64 *pfilesize);
@@ -875,12 +838,16 @@ public:
 	static ProbeResult ProbeFileHeaderPSM16(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderPT36(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderPTM(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderPuma(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderRTM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderS3M(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderSFX(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderSTM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderSTP(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderSTX(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderSymMOD(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderTCB(MemoryFileReader file, const uint64 *pfilesize);
+	static ProbeResult ProbeFileHeaderUNIC(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderULT(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderXM(MemoryFileReader file, const uint64 *pfilesize);
 	static ProbeResult ProbeFileHeaderXMF(MemoryFileReader file, const uint64 *pfilesize);
@@ -898,6 +865,7 @@ public:
 	bool ReadAMS(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadAMS2(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadC67(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadCBA(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadDBM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadDTM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadDIGI(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
@@ -905,18 +873,24 @@ public:
 	bool ReadDSM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadDSm(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadDSym(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadETX(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadFAR(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadFC(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadFMT(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadFTM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadGDM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadGMC(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadGT2(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadGTK(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadICE(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadIMF(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadIMS(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadIT(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadITP(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadJ2B(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadKRIS(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadMUS_KM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
-	bool ReadM15(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadSTK(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadMDL(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadMED(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadMO3(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
@@ -929,12 +903,16 @@ public:
 	bool ReadPSM16(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadPT36(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadPTM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadPuma(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadRTM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadS3M(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadSFX(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadSTM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadSTP(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadSTX(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadSymMOD(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadTCB(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
+	bool ReadUNIC(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadULT(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadXM(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
 	bool ReadXMF(FileReader &file, ModLoadingFlags loadFlags = loadCompleteModule);
@@ -958,18 +936,21 @@ public:
 	bool SaveMod(std::ostream &f) const;
 	bool SaveIT(std::ostream &f, const mpt::PathString &filename, bool compatibilityExport = false);
 	uint32 SaveMixPlugins(std::ostream *file=nullptr, bool bUpdate=true);
-	void WriteInstrumentPropertyForAllInstruments(uint32 code,  uint16 size, std::ostream &f, INSTRUMENTINDEX nInstruments) const;
-	void SaveExtendedInstrumentProperties(INSTRUMENTINDEX nInstruments, std::ostream &f) const;
+	void SaveExtendedInstrumentProperties(INSTRUMENTINDEX instr, MODTYPE forceType, std::ostream &f) const;
+	static void SaveExtendedInstrumentProperties(mpt::span<const ModInstrument * const> instruments, MODTYPE forceType, std::ostream &f, bool allInstruments);
 	void SaveExtendedSongProperties(std::ostream &f) const;
 #endif // MODPLUG_NO_FILESAVE
+	static void ReadExtendedInstrumentProperty(mpt::span<ModInstrument *> instruments, const uint32 code, FileReader &file);
 	bool LoadExtendedSongProperties(FileReader &file, bool ignoreChannelCount, bool* pInterpretMptMade = nullptr);
 	void LoadMPTMProperties(FileReader &file, uint16 cwtv);
 
+	static mpt::ustring GetImpulseTrackerVersion(uint16 cwtv, uint16 cmwt);
 	static mpt::ustring GetSchismTrackerVersion(uint16 cwtv, uint32 reserved);
 
-	// Reads extended instrument properties(XM/IT/MPTM).
+	// Reads extended instrument properties(XM/IT/MPTM/ITI/XI).
 	// Returns true if extended instrument properties were found.
-	bool LoadExtendedInstrumentProperties(FileReader &file);
+	bool LoadExtendedInstrumentProperties(FileReader &file) { return LoadExtendedInstrumentProperties(mpt::as_span(Instruments).subspan(1, GetNumInstruments()), file); }
+	static bool LoadExtendedInstrumentProperties(mpt::span<ModInstrument *> instruments, FileReader &file);
 
 	void SetDefaultPlaybackBehaviour(MODTYPE type);
 	static PlayBehaviourSet GetSupportedPlaybackBehaviour(MODTYPE type);
@@ -986,12 +967,13 @@ public:
 	// Converts 4 bytes formatted like SoundTracker/NoiseTracker/ProTracker pattern data by converting the period to a note and filling the instrument number, and returns the effect command and parameter bytes.
 	static std::pair<uint8, uint8> ReadMODPatternEntry(const std::array<uint8, 4> data, ModCommand &m);
 
-	void SetupMODPanning(bool bForceSetup = false); // Setup LRRL panning, max channel volume
+	void SetupMODPanning(bool forceSetup = false);  // Setup LRRL panning
 
 public:
 	// Real-time sound functions
 	void SuspendPlugins();
 	void ResumePlugins();
+	void UpdatePluginPositions();
 	void StopAllVsti();
 	void RecalculateGainForAllPlugs();
 	void ResetChannels();
@@ -1006,6 +988,8 @@ public:
 	samplecount_t ReadOneTick();
 private:
 	void CreateStereoMix(int count);
+	bool MixChannel(int count, ModChannel &chn, CHANNELINDEX channel, bool doMix);
+	std::pair<mixsample_t *, mixsample_t *> GetChannelOffsets(const ModChannel &chn, CHANNELINDEX channel);
 public:
 	bool FadeSong(uint32 msec);
 private:
@@ -1014,7 +998,7 @@ private:
 	void ProcessInputChannels(IAudioSource &source, std::size_t countChunk);
 public:
 	samplecount_t GetTotalSampleCount() const { return m_PlayState.m_lTotalSampleCount; }
-	bool HasPositionChanged() { bool b = m_PlayState.m_bPositionChanged; m_PlayState.m_bPositionChanged = false; return b; }
+	bool HasPositionChanged() { bool b = m_PlayState.m_flags[SONG_POSITIONCHANGED]; m_PlayState.m_flags.reset(SONG_POSITIONCHANGED); return b; }
 	bool IsRenderingToDisc() const { return m_bIsRendering; }
 
 	void PrecomputeSampleLoops(bool updateChannels = false);
@@ -1038,6 +1022,7 @@ public:
 	void SetupNextRow(PlayState &playState, const bool patternLoop) const;
 	CHANNELINDEX GetNNAChannel(CHANNELINDEX nChn) const;
 	CHANNELINDEX CheckNNA(CHANNELINDEX nChn, uint32 instr, int note, bool forceCut);
+	void StopOldNNA(ModChannel &chn, CHANNELINDEX channel);
 	void NoteChange(ModChannel &chn, int note, bool bPorta = false, bool bResetEnv = true, bool bManual = false, CHANNELINDEX channelHint = CHANNELINDEX_INVALID) const;
 	void InstrumentChange(ModChannel &chn, uint32 instr, bool bPorta = false, bool bUpdVol = true, bool bResetEnv = true) const;
 	void ApplyInstrumentPanning(ModChannel &chn, const ModInstrument *instr, const ModSample *smp) const;
@@ -1046,7 +1031,8 @@ public:
 	// Channel Effects
 	void KeyOff(ModChannel &chn) const;
 	// Global Effects
-	void SetTempo(TEMPO param, bool setAsNonModcommand = false);
+	void SetTempo(TEMPO param, bool setAsNonModcommand = false) { SetTempo(m_PlayState, param, setAsNonModcommand); }
+	void SetTempo(PlayState &playState, TEMPO param, bool setAsNonModcommand = false) const;
 	void SetSpeed(PlayState &playState, uint32 param) const;
 	static TEMPO ConvertST2Tempo(uint8 tempo);
 
@@ -1057,8 +1043,7 @@ public:
 protected:
 	// Global variable initializer for loader functions
 	void SetType(MODTYPE type);
-	void InitializeGlobals(MODTYPE type = MOD_TYPE_NONE);
-	void InitializeChannels();
+	void InitializeGlobals(MODTYPE type, CHANNELINDEX numChannels);
 
 	// Channel effect processing
 	int GetVibratoDelta(int type, int position) const;
@@ -1096,7 +1081,10 @@ protected:
 		Pan8bit = 8,
 	};
 	// Channel Effects
+	void ResetAutoSlides(ModChannel &chn) const;
+	void ProcessAutoSlides(PlayState &playState, CHANNELINDEX channel);
 	void UpdateS3MEffectMemory(ModChannel &chn, ModCommand::PARAM param) const;
+	void PortamentoFC(ModChannel &chn) const;
 	void PortamentoUp(CHANNELINDEX nChn, ModCommand::PARAM param, const bool doFinePortamentoAsRegular);
 	void PortamentoUp(PlayState &playState, CHANNELINDEX nChn, ModCommand::PARAM param, const bool doFinePortamentoAsRegular) const;
 	void PortamentoDown(CHANNELINDEX nChn, ModCommand::PARAM param, const bool doFinePortamentoAsRegular);
@@ -1113,13 +1101,19 @@ protected:
 	int16 CalculateFinetuneTarget(PATTERNINDEX pattern, ROWINDEX row, CHANNELINDEX channel) const;
 	void NoteSlide(ModChannel &chn, uint32 param, bool slideUp, bool retrig) const;
 	std::pair<uint16, bool> GetVolCmdTonePorta(const ModCommand &m, uint32 startTick) const;
+	bool TonePortamentoSharesEffectMemory() const;
+	void InitTonePortamento(ModChannel &chn, uint16 param) const;
 	void TonePortamento(CHANNELINDEX chn, uint16 param);
 	int32 TonePortamento(PlayState &playState, CHANNELINDEX nChn, uint16 param) const;
+	void TonePortamentoWithDuration(ModChannel &chn, uint16 param = uint16_max) const;
 	void Vibrato(ModChannel &chn, uint32 param) const;
 	void FineVibrato(ModChannel &chn, uint32 param) const;
-	void VolumeSlide(ModChannel &chn, ModCommand::PARAM param) const;
+	void AutoVolumeSlide(ModChannel &chn, ModCommand::PARAM param) const;
+	void VolumeDownETX(const PlayState &playState, ModChannel &chn, ModCommand::PARAM param) const;
+	void VolumeSlide(ModChannel &chn, ModCommand::PARAM param, bool volCol = false) const;
 	void PanningSlide(ModChannel &chn, ModCommand::PARAM param, bool memory = true) const;
 	void ChannelVolSlide(ModChannel &chn, ModCommand::PARAM param) const;
+	void ChannelVolumeDownWithDuration(ModChannel &chn, uint16 param = uint16_max) const;
 	void FineVolumeUp(ModChannel &chn, ModCommand::PARAM param, bool volCol) const;
 	void FineVolumeDown(ModChannel &chn, ModCommand::PARAM param, bool volCol) const;
 	void Tremolo(ModChannel &chn, uint32 param) const;
@@ -1136,27 +1130,25 @@ protected:
 	bool HandleNextRow(PlayState &state, const ModSequence &order, bool honorPatternLoop) const;
 	void ExtendedMODCommands(CHANNELINDEX nChn, ModCommand::PARAM param);
 	void ExtendedS3MCommands(CHANNELINDEX nChn, ModCommand::PARAM param);
-	void ExtendedChannelEffect(ModChannel &chn, uint32 param);
+	void ExtendedChannelEffect(ModChannel &chn, uint32 param, PlayState &playState) const;
 	void InvertLoop(ModChannel &chn);
 	void PositionJump(PlayState &state, CHANNELINDEX chn) const;
 	ROWINDEX PatternBreak(PlayState &state, CHANNELINDEX chn, uint8 param) const;
-	void GlobalVolSlide(ModCommand::PARAM param, uint8 &nOldGlobalVolSlide);
+	void GlobalVolSlide(PlayState &playState, ModCommand::PARAM param, CHANNELINDEX chn) const;
 
 	void ProcessMacroOnChannel(CHANNELINDEX nChn);
 	void ProcessMIDIMacro(PlayState &playState, CHANNELINDEX nChn, bool isSmooth, const MIDIMacroConfigData::Macro &macro, uint8 param = 0, PLUGINDEX plugin = 0);
-	void ParseMIDIMacro(PlayState &playState, CHANNELINDEX nChn, bool isSmooth, const mpt::span<const char> macro, mpt::span<uint8> &out, uint8 param = 0, PLUGINDEX plugin = 0) const;
-	static float CalculateSmoothParamChange(const PlayState &playState, float currentValue, float param);
 	void SendMIDIData(PlayState &playState, CHANNELINDEX nChn, bool isSmooth, const mpt::span<const uint8> macro, PLUGINDEX plugin);
-	void SendMIDINote(CHANNELINDEX chn, uint16 note, uint16 volume);
+	void SendMIDINote(CHANNELINDEX chn, uint16 note, uint16 volume, IMixPlugin *plugin = nullptr);
 
 	int SetupChannelFilter(ModChannel &chn, bool bReset, int envModifier = 256) const;
 	int HandleNoteChangeFilter(ModChannel &chn) const;
 
-	// Low-Level effect processing
-	void DoFreqSlide(ModChannel &chn, int32 &period, int32 amount, bool isTonePorta = false) const;
-	void UpdateTimeSignature();
-
 public:
+	static float CalculateSmoothParamChange(const PlayState &playState, float currentValue, float param);
+
+	void DoFreqSlide(ModChannel &chn, int32 &period, int32 amount, bool isTonePorta = false) const;
+
 	// Convert frequency to IT cutoff (0...127)
 	uint8 FrequencyToCutOff(double frequency) const;
 	// Convert IT cutoff (0...127 + modifier) to frequency
@@ -1188,6 +1180,11 @@ public:
 		return UseCombinedPortamentoCommands(GetType());
 	}
 
+	uint32 GlobalVolumeRange() const noexcept
+	{
+		return !(GetType() & (MOD_TYPE_IT | MOD_TYPE_MPT | MOD_TYPE_IMF | MOD_TYPE_J2B | MOD_TYPE_MID | MOD_TYPE_AMS | MOD_TYPE_DBM | MOD_TYPE_PTM | MOD_TYPE_MDL | MOD_TYPE_DTM)) ? 64 : 128;
+	}
+
 	bool DestroySample(SAMPLEINDEX nSample);
 	bool DestroySampleThreadsafe(SAMPLEINDEX nSample);
 
@@ -1223,7 +1220,7 @@ protected:
 	bool ReadXISample(SAMPLEINDEX nSample, FileReader &file);
 	bool ReadITSSample(SAMPLEINDEX nSample, FileReader &file, bool rewind = true);
 	bool ReadITISample(SAMPLEINDEX nSample, FileReader &file);
-	bool ReadIFFSample(SAMPLEINDEX sample, FileReader &file, bool allowLittleEndian = true);
+	bool ReadIFFSample(SAMPLEINDEX sample, FileReader &file, bool allowLittleEndian = true, uint8 octave = uint8_max);
 	bool ReadBRRSample(SAMPLEINDEX sample, FileReader &file);
 	bool ReadFLACSample(SAMPLEINDEX sample, FileReader &file);
 	bool ReadOpusSample(SAMPLEINDEX sample, FileReader &file);
@@ -1272,9 +1269,9 @@ public:
 	// Resolve note/instrument combination to real sample index. Return value is guaranteed to be in [0, GetNumSamples()].
 	SAMPLEINDEX GetSampleIndex(ModCommand::NOTE note, uint32 instr) const noexcept;
 
-	uint32 MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChannel, uint8 note, bool isXG, std::bitset<16> drumChns);
+	uint32 MapMidiInstrument(uint8 program, uint16 bank, uint8 midiChannel, uint8 note, bool isXG, std::bitset<32> drumChns);
 	size_t ITInstrToMPT(FileReader &file, ModInstrument &ins, uint16 trkvers);
-	std::pair<bool, bool> LoadMixPlugins(FileReader &file);
+	std::pair<bool, bool> LoadMixPlugins(FileReader &file, bool ignoreChannelCount = true);
 #ifndef NO_PLUGINS
 	static void ReadMixPluginChunk(FileReader &file, SNDMIXPLUGIN &plugin);
 	void ProcessMidiOut(CHANNELINDEX nChn);
@@ -1284,12 +1281,16 @@ public:
 	void ProcessStereoSeparation(samplecount_t countChunk);
 
 private:
-	PLUGINDEX GetChannelPlugin(const PlayState &playState, CHANNELINDEX nChn, PluginMutePriority respectMutes) const;
+	PLUGINDEX GetChannelPlugin(const ModChannel &channel, CHANNELINDEX nChn, PluginMutePriority respectMutes) const;
 	static PLUGINDEX GetActiveInstrumentPlugin(const ModChannel &chn, PluginMutePriority respectMutes);
 	IMixPlugin *GetChannelInstrumentPlugin(const ModChannel &chn) const;
 
 public:
-	PLUGINDEX GetBestPlugin(const PlayState &playState, CHANNELINDEX nChn, PluginPriority priority, PluginMutePriority respectMutes) const;
+	PLUGINDEX GetBestPlugin(const ModChannel &channel, CHANNELINDEX nChn, PluginPriority priority, PluginMutePriority respectMutes) const;
+
+#if defined(MPT_ENABLE_PLAYBACK_TRACE)
+	PlaybackTest CreatePlaybackTest(PlaybackTestSettings settings);
+#endif // MPT_ENABLE_PLAYBACK_TRACE
 
 };
 
@@ -1305,22 +1306,10 @@ inline IMixPlugin* CSoundFile::GetInstrumentPlugin(INSTRUMENTINDEX instr) const
 #endif // NO_PLUGINS
 
 
-///////////////////////////////////////////////////////////
-// Low-level Mixing functions
-
 #define FADESONGDELAY		100
 
 MPT_CONSTEXPRINLINE int8 MOD2XMFineTune(int v) { return static_cast<int8>(static_cast<uint8>(v) << 4); }
 MPT_CONSTEXPRINLINE int8 XM2MODFineTune(int v) { return static_cast<int8>(static_cast<uint8>(v) >> 4); }
 
-// Read instrument property with 'code' and 'size' from 'file' to instrument 'pIns'.
-void ReadInstrumentExtensionField(ModInstrument* pIns, const uint32 code, const uint16 size, FileReader &file);
-
-// Read instrument property with 'code' from 'file' to instrument 'pIns'.
-void ReadExtendedInstrumentProperty(ModInstrument* pIns, const uint32 code, FileReader &file);
-
-// Read extended instrument properties from 'file' to instrument 'pIns'.
-void ReadExtendedInstrumentProperties(ModInstrument* pIns, FileReader &file);
-
 
 OPENMPT_NAMESPACE_END
